"""AMPQ support."""
import asyncio
import pickle
import uuid
from importlib import import_module

import aioamqp

from . import logger, AIOFALSE
from .utils import AsyncMixin, AttrDict


class Queue(AsyncMixin):

    defaults = dict(
        host='localhost',
        port=None,
        login='guest',
        password='guest',
        virtualhost='/',
    )

    def __init__(self, core, loop=None, exchange='donald', queue='donald', **params):
        """Initialize the queue."""
        self.params = self.defaults
        self.params.update(params)

        self._core = core
        self._loop = loop or asyncio.get_event_loop()
        self._queue = queue
        self._transport = None
        self._protocol = None
        self._channel = None
        self._started = False
        self._connected = False

    def is_connected(self):
        """Check that the queue is connected."""
        return self._connected

    def start(self, listen=True):
        """Connect to message queue."""
        if self._core.params.always_eager:
            return False
        self._started = True
        return asyncio.ensure_future(self.connect(listen), loop=self.loop)

    @asyncio.coroutine
    def connect(self, listen=True):
        """Connect to queue."""
        logger.warn('Connect to queue.')
        try:
            self._transport, self._protocol = yield from aioamqp.connect(
                loop=self._loop, on_error=self.on_error, **self.params)
            self._channel = yield from self._protocol.channel()

            yield from self._channel.queue_declare(queue_name=self._queue, durable=True)
            yield from self._channel.basic_qos(
                prefetch_count=1, prefetch_size=0, connection_global=False)
            self._connected = True
            if listen:
                yield from self.listen()

        except (aioamqp.AmqpClosedConnection, OSError) as exc:
            self.on_error(exc)

    def on_error(self, exc):
        """Error's callback."""
        self._connected = False
        if not self._started or self.is_closed():
            return False

        logger.error(exc)
        self.loop.call_later(1, asyncio.ensure_future, self.connect())

    def listen(self):
        """Run tasks from self.queue.

        :returns: A coroutine
        """
        if self._core.params.always_eager:
            return AIOFALSE
        return self._channel.basic_consume(self.callback, queue_name=self._queue)

    @asyncio.coroutine
    def stop(self):
        """Stop listeners."""
        if not self.is_connected() or self.is_closed():
            return False

        logger.warn('Disconnect from queue.')
        self._started = False
        yield from self._protocol.close()
        self._transport.close()
        self._connected = False

    def submit(self, func, *args, **kwargs):
        """Submit to the queue."""
        logger.info('Submit task to queue.')

        payload = pickle.dumps((func, args, kwargs))
        properties = dict(delivery_mode=2, message_id=str(uuid.uuid4()))

        if self._core.params.always_eager:
            return asyncio.ensure_future(
                self.callback(self._channel, payload, None, AttrDict(properties)), loop=self.loop)

        if asyncio.iscoroutine(func):
            raise RuntimeError('Submit coroutines to queue as coroutine-functions with params.')

        coro = self._channel.basic_publish(
            payload=payload, exchange_name='', routing_key=self._queue, properties=properties
        )
        return asyncio.ensure_future(coro, loop=self.loop)

    @asyncio.coroutine
    def callback(self, channel, body, envelope, properties):
        """A Listener."""
        func, args, kwargs = pickle.loads(body)
        if isinstance(func, str):
            mod, _, func = func.rpartition('.')
            try:
                mod = import_module(mod)
                func = getattr(mod, func)
            except (ImportError, AttributeError) as exc:
                logger.error(exc)
                return False

        try:
            result = yield from self._core.submit(func, *args, **kwargs)
        except Exception as exc:
            logger.error('Get exception for %r', properties.message_id)
            return exc

        logger.info('Get result %r %r', result, properties.message_id)
        if channel:
            yield from channel.basic_client_ack(delivery_tag=envelope.delivery_tag)
        return result



"""Donald tasks' manager for Asyncio."""
import logging
import asyncio


AIOFALSE = asyncio.Future()
AIOFALSE.set_result(False)


logger = logging.getLogger(__name__)
handle = logging.StreamHandler()
handle.setFormatter(logging.Formatter('%(thread)d: %(message)s'))
logger.addHandler(handle)


from .core import Donald # noqa


__version__ = "0.0.17"
__project__ = "Donald"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "BSD"



import asyncio
import threading

from . import logger
from .utils import AsyncMixin, CallableFuture, Future


class AsyncThreadWorker(AsyncMixin, threading.Thread):

    def __init__(self, *args, **kwargs):
        super(AsyncThreadWorker, self).__init__(*args, **kwargs)
        self._loop = None

    def start(self):
        self._future = Future()
        super(AsyncThreadWorker, self).start()
        return self._future

    def run(self):
        logger.info('Start loop for worker: %d' % id(self))
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self._loop.call_soon_threadsafe(
            asyncio.futures._set_result_unless_cancelled, self._future, True)
        self._loop.run_forever()

    def stop(self):
        if self.is_closed():
            fut = Future()
            fut.set_result(False)
            return fut
        logger.info('Stop loop for worker: %d' % id(self))
        for task in asyncio.Task.all_tasks(loop=self._loop):
            task.cancel()

        fut = CallableFuture(self._loop.stop)
        self._loop.call_soon_threadsafe(fut)
        return fut

    def submit(self, func, *args, **kwargs):
        """Submit function/coroutine to current loop."""
        if not self.is_running():
            raise RuntimeError('Worker loop is stopped.')

        logger.info('Submit task with worker: %d' % id(self))
        job = call_with_loop(self._loop, func, *args, **kwargs)
        waiter = Future()
        self._loop.call_soon(
            asyncio.futures._set_result_unless_cancelled, waiter, True)
        return job, waiter


def call_with_loop(loop, func, *args, **kwargs):
    """Call given coro/function with given loop."""
    if asyncio.iscoroutine(func):
        return asyncio.run_coroutine_threadsafe(func, loop)

    if asyncio.iscoroutinefunction(func):
        return asyncio.run_coroutine_threadsafe(func(*args, **kwargs), loop)

    fut = CallableFuture(func, *args, **kwargs)
    loop.call_soon_threadsafe(fut)
    return fut



import os
from concurrent.futures import Future


class AsyncMixin:

    def is_running(self):
        return self._loop and self._loop.is_running()

    def is_closed(self):
        return not self._loop or self._loop.is_closed()

    @property
    def loop(self):
        return self._loop


class AttrDict(dict):

    def __init__(self, *args, **kw):
        self.__dict__ = self
        super(AttrDict, self).__init__(*args, **kw)


class CallableFuture(Future):

    def __init__(self, func, *args, **kw):
        super(CallableFuture, self).__init__()
        self._func = func
        self._args = args
        self._kwargs = kw

    def __call__(self):
        with self._condition:
            try:
                self.set_result(self._func(*self._args, **self._kwargs))
            except Exception as exc:
                self.set_exception(exc)


class Singleton(type):

    instance = None

    def __call__(cls, *args, **kw):
        if not cls.instance:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


class FileLocked(Exception):
    pass


class FileLock(object):

    """Simplest filelock implementation."""

    def __init__(self, fname, timeout=None, force=False):
        self.fname = fname
        self.fh = None
        self.flags = os.O_CREAT | os.O_RDWR
        for flag in ('O_EXLOCK', 'O_NOINHERIT'):
            self.flags |= getattr(os, flag, 0)

    def acquire(self):
        if os.path.exists(self.fname):
            raise FileLocked()
        self.fh = os.open(self.fname, self.flags)
        os.write(self.fh, bytes(os.getpid()))

    def release(self):
        if self.fh:
            os.close(self.fh)
        os.remove(self.fname)

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.release()



import click
import asyncio
from donald import Donald
from importlib import import_module


@asyncio.coroutine
def run_donald(donald, listen=None, module=None):
    yield from donald.start()
    if module:
        try:
            click.echo('Import module: %s' % module)
            import_module(module)
        except ImportError:
            click.echo('Invalid module: %s' % module)


@click.command()
@click.option('--listen', help="Listen the given AMPQ URL.")
@click.option('--workers', default=4, help="Number of workers.")
@click.option('--loglevel', default='INFO', help="Logging level.")
@click.option('--module', help="Load module.")
def main(workers=4, loglevel='info', **kw):
    """Run Donald server."""
    donald = Donald(num_threads=workers, loglevel=loglevel)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_donald(donald, **kw))
    loop.run_forever()


if __name__ == '__main__':
    main()



"""Main Donald class."""

import asyncio
import atexit
import datetime
import random

from . import logger, AIOFALSE
from .queue import Queue
from .utils import AsyncMixin, AttrDict, Singleton, FileLock, FileLocked
from .worker import AsyncThreadWorker, call_with_loop


class Donald(AsyncMixin, metaclass=Singleton):

    """I'am on Donald."""

    defaults = dict(

        # Run tasks imediatelly
        always_eager=False,

        # Number of workers
        num_threads=4,

        # Ensure that the Donald starts only once
        filelock=None,

        # logging level
        loglevel='INFO',

        # AMQP params
        queue=dict(
            exchange='donald',
            queue='donald',
        ),

    )

    def __init__(self, loop=None, **params):
        """Initialize donald parameters."""
        self.params = AttrDict(self.defaults)
        self.params.update(params)

        logger.setLevel(self.params['loglevel'].upper())
        logger.propagate = False

        # Base loop and queue
        self._loop = loop or asyncio.get_event_loop()
        num_threads = 1 if self.params.num_threads < 1 else self.params.num_threads
        self._threads = tuple(AsyncThreadWorker() for _ in range(num_threads))
        self._waiters = {}
        self._schedules = []
        self._closing = False
        self._started = False
        self._lock = FileLock(self.params.filelock)

        self.queue = Queue(self, loop=self._loop, **self.params.queue)
        self.params.always_eager = self.params.always_eager or not len(self._threads)

    def start(self, loop=None):
        """Start workers.

        :returns: A coroutine
        """
        logger.warn('Start Donald')
        if loop is not None:
            self._loop = self.queue._loop = loop

        if self.params.filelock:
            try:
                self._lock.acquire()
            except FileLocked:
                logger.warn('Donald is locked. Exit.')
                return AIOFALSE

        self._closing = False
        atexit.register(self.stop)

        def start_worker(w):
            return self.future(w.start())

        self._started = True
        return asyncio.wait(map(start_worker, self._threads))

    def stop(self, *args):
        """Stop workers. Disconnect from queue. Cancel schedules.

        The method could be called syncroniously too.

        :returns: A future
        """
        # Stop queue
        if self.is_closed():
            return AIOFALSE

        tasks = [asyncio.ensure_future(self.queue.stop(), loop=self._loop)]

        if self._closing or not self._started:
            return asyncio.wait(tasks, loop=self._loop)

        logger.warn('Stop Donald.')

        if self.params.filelock:
            self._lock.release()

        for task in self._schedules:
            task.cancel()

        # Stop workers
        tasks += [
            asyncio.ensure_future(self.future(t.stop()), loop=self._loop)
            for t in self._threads]

        self._closing = True

        return asyncio.wait(tasks, loop=self._loop)

    def future(self, fut):
        """Just a shortcut on asyncio.wrap_future."""
        return asyncio.wrap_future(fut, loop=self._loop)

    def submit(self, func, *args, **kwargs):
        """Submit given function/coroutine."""
        if self.params.always_eager:
            future = call_with_loop(self._loop, func, *args, **kwargs)
            return self.future(future)

        return asyncio.ensure_future(self._submit(func, *args, **kwargs))

    @asyncio.coroutine
    def _submit(self, func, *args, **kwargs):
        """Wait for free worker and submit to it."""
        if len(self._waiters) == len(self._threads):
            yield from asyncio.wait(
                self._waiters, loop=self._loop, return_when=asyncio.FIRST_COMPLETED)
        worker = random.choice(self._threads)
        future, waiter = map(self.future, worker.submit(func, *args, **kwargs))
        self._waiters[id(worker)] = waiter
        waiter.add_done_callback(lambda f: self._waiters.pop(id(worker), None))
        return (yield from future)

    def schedule(self, interval, func, *args, **kwargs):
        """Run given func/coro periodically."""
        if isinstance(interval, datetime.timedelta):
            interval = interval.total_seconds()

        if not isinstance(interval, float):
            interval = float(interval)

        @asyncio.coroutine
        def scheduler():
            while self.is_running():
                self.submit(func, *args, **kwargs)
                yield from asyncio.sleep(interval, loop=self._loop)

        logger.info('Schedule %r' % func)
        self._schedules.append(asyncio.ensure_future(scheduler(), loop=self._loop))

    def __str__(self):
        """String representation."""
        return 'Donald [%s]' % self.params['num_threads']
